"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import os
import pandas as pd
import pathlib as pl
import datetime as dt

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QTextCursor
from PyQt5.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QComboBox,
    QPushButton,
    QMessageBox,
    QSpacerItem,
    QSizePolicy,
    QFileDialog,
    QTableWidgetItem,
    QApplication,
    QMenu,
    QListWidgetItem,
)

from . import config as cfg
from . import dialog, plot_data_module, project_functions
from . import utilities as util
from . import gui_utilities
from .observation_ui import Ui_Form


class AssignConverter(QDialog):
    """
    dialog for assigning converter to selected column
    """

    def __init__(self, columns, converters, col_conv):
        super().__init__()

        self.setWindowTitle("Converters")

        self.vbox = QVBoxLayout()

        self.label = QLabel()
        self.label.setText("Assign converter to column")
        self.vbox.addWidget(self.label)

        self.cbb = []
        for column_idx in columns.split(","):
            hbox = QHBoxLayout()
            hbox.addWidget(QLabel(f"Column #{column_idx}:"))
            self.cbb.append(QComboBox())
            self.cbb[-1].addItems(["None"] + sorted(converters.keys()))

            if column_idx in col_conv:
                self.cbb[-1].setCurrentIndex((["None"] + sorted(converters.keys())).index(col_conv[column_idx]))
            else:
                self.cbb[-1].setCurrentIndex(0)
            hbox.addWidget(self.cbb[-1])
            self.vbox.addLayout(hbox)

        hbox1 = QHBoxLayout()
        self.pbOK = QPushButton("OK")
        self.pbOK.clicked.connect(self.accept)
        self.pbCancel = QPushButton("Cancel")
        self.pbCancel.clicked.connect(self.reject)
        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        hbox1.addItem(spacerItem)
        hbox1.addWidget(self.pbCancel)
        hbox1.addWidget(self.pbOK)
        self.vbox.addLayout(hbox1)

        self.setLayout(self.vbox)


class Observation(QDialog, Ui_Form):
    def __init__(self, tmp_dir: str, project_path: str = "", converters: dict = {}, time_format: str = cfg.S, parent=None):
        """
        Args:
            tmp_dir (str): path of temporary directory
            project_path (str): path of project
            converters (dict): converters dictionary
        """

        super().__init__()

        self.tmp_dir = tmp_dir
        self.project_path = project_path
        self.converters = converters
        self.time_format = time_format
        self.observation_time_interval: tuple = [0, 0]
        self.mem_dir = ""
        self.test = None

        self.setupUi(self)

        # insert duration widget for time offset
        # self.obs_time_offset = duration_widget.Duration_widget(0)
        self.obs_time_offset = dialog.get_time_widget(0)
        self.horizontalLayout_6.insertWidget(1, self.obs_time_offset)
        self.obs_time_offset.setEnabled(False)

        # time offset
        self.cb_time_offset.stateChanged.connect(self.cb_time_offset_changed)
        # date offset
        """self.cb_date_offset.stateChanged.connect(self.cb_date_offset_changed)"""

        # observation type
        self.rb_media_files.toggled.connect(self.obs_type_changed)
        self.rb_live.toggled.connect(self.obs_type_changed)
        self.rb_images.toggled.connect(self.obs_type_changed)

        # button menu for media
        menu_items = [
            "media abs path|with absolute path",
            "media rel path|with relative path",
            {
                "from directory": [
                    "dir abs path|with absolute path ",
                    "dir rel path|wih relative path ",
                ]
            },
        ]
        menu = QMenu()
        menu.triggered.connect(lambda x: self.add_media(mode=x.statusTip()))
        self.add_button_menu(menu_items, menu)
        self.pbAddVideo.setMenu(menu)

        self.pbRemoveVideo.clicked.connect(self.remove_media)

        # button menu for data file
        data_menu_items = [
            "data abs path|with absolute path",
            "data rel path|with relative path",
        ]

        menu_data = QMenu()
        menu_data.triggered.connect(lambda x: self.add_data_file(mode=x.statusTip()))
        self.add_button_menu(data_menu_items, menu_data)
        self.pb_add_data_file.setMenu(menu_data)

        # button menu for images
        images_menu_items = [
            "images abs path|with absolute path",
            "images rel path|with relative path",
        ]

        menu_images = QMenu()
        menu_images.triggered.connect(lambda x: self.add_images_directory(mode=x.statusTip()))
        self.add_button_menu(images_menu_items, menu_images)
        self.pb_add_directory.setMenu(menu_images)

        self.pb_remove_data_file.clicked.connect(self.remove_data_file)
        self.pb_view_data_head.clicked.connect(self.view_data_file_head_tail)
        self.pb_plot_data.clicked.connect(self.plot_data_file)

        self.pb_use_media_file_name_as_obsid.clicked.connect(self.use_media_file_name_as_obsid)
        self.pb_use_img_dir_as_obsid.clicked.connect(self.use_img_dir_as_obsid)

        self.cbVisualizeSpectrogram.clicked.connect(self.extract_wav)
        self.cb_visualize_waveform.clicked.connect(self.extract_wav)

        # self.cb_media_creation_date_as_offset.clicked.connect(self.check_creation_date)

        self.cb_observation_time_interval.clicked.connect(self.limit_time_interval)

        self.pbSave.clicked.connect(self.pbSave_clicked)
        self.pbLaunch.clicked.connect(self.pbLaunch_clicked)
        self.pbCancel.clicked.connect(self.pbCancel_clicked)

        self.tw_data_files.cellDoubleClicked[int, int].connect(self.tw_data_files_cellDoubleClicked)

        self.mediaDurations, self.mediaFPS, self.mediaHasVideo, self.mediaHasAudio, self.media_creation_time = {}, {}, {}, {}, {}

        for w in (
            self.cbVisualizeSpectrogram,
            self.cb_visualize_waveform,
            self.cb_observation_time_interval,
            self.cb_media_creation_date_as_offset,
            self.cbCloseCurrentBehaviorsBetweenVideo,
        ):
            w.setEnabled(False)

        self.cb_observation_time_interval.setEnabled(True)

        self.cb_start_from_current_time.stateChanged.connect(self.cb_start_from_current_time_changed)

        # images
        # self.pb_add_directory.clicked.connect(self.add_images_directory)
        self.pb_remove_directory.clicked.connect(self.remove_images_directory)

        self.tabWidget.setCurrentIndex(0)

        # geometry
        gui_utilities.restore_geometry(self, "new observation", (800, 650))

    # def cb_date_offset_changed(self):
    #    """
    #    activate/desactivate time value
    #    """
    #    self.de_date_offset.setEnabled(self.cb_date_offset.isChecked())

    def cb_time_offset_changed(self):
        """
        activate/desactivate date value
        """
        self.obs_time_offset.setEnabled(self.cb_time_offset.isChecked())

    def use_media_file_name_as_obsid(self) -> None:
        """
        set observation id with the media file name value (without path)
        """
        if not self.twVideo1.rowCount():
            QMessageBox.critical(self, cfg.programName, "A media file must be loaded in player #1")
            return

        first_media_file = ""
        for row in range(self.twVideo1.rowCount()):
            if int(self.twVideo1.cellWidget(row, 0).currentText()) == 1:
                first_media_file = self.twVideo1.item(row, 2).text()
                break
        # check if player #1 is used
        if not first_media_file:
            QMessageBox.critical(self, cfg.programName, "A media file must be loaded in player #1")
            return

        self.leObservationId.setText(pl.Path(first_media_file).name)

    def use_img_dir_as_obsid(self) -> None:
        """
        set observation id with the images directory (without path)
        """

        if not self.lw_images_directory.count():
            QMessageBox.critical(self, cfg.programName, "You have to select at least one images directory")
            return

        self.leObservationId.setText(pl.Path(self.lw_images_directory.item(0).text()).name)

    def obs_type_changed(self) -> None:
        """
        change stacked widget page in base at the observation type
        """

        for idx, rb in enumerate((self.rb_media_files, self.rb_live, self.rb_images)):
            if rb.isChecked():
                self.sw_observation_type.setCurrentIndex(idx + 1)

        # hide 'limit observation to time interval' for images
        self.cb_observation_time_interval.setEnabled(not self.rb_images.isChecked())

    def add_images_directory(self, mode: str):
        """
        add path to images directory
        """

        if mode.split("|")[0] not in (
            "images abs path",
            "images rel path",
        ):
            QMessageBox.critical(
                self,
                cfg.programName,
                (f"Wrong mode to add a pictures directory {mode}"),
            )
            return

        # check if project saved
        if (" w/o" in mode or " rel " in mode) and (not self.project_file_name):
            QMessageBox.critical(
                self,
                cfg.programName,
                ("It is not possible to add a pictures directory with a relative path if the project is not already saved"),
            )
            return

        fd = QFileDialog()
        fd.setDirectory(os.path.expanduser("~") if (" abs " in mode) else str(pl.Path(self.project_path).parent))

        dir_path = fd.getExistingDirectory(None, "Select directory")

        if not dir_path:
            return

        result = util.dir_images_number(dir_path)
        if not result.get("number of images", 0):
            response = dialog.MessageDialog(
                cfg.programName,
                f"The directory does not contain images ({','.join(cfg.IMAGE_EXTENSIONS)})",
                ["Cancel", "Add directory"],
            )
            if response == "Cancel":
                return

        # store directory for next usage
        self.mem_dir = str(pl.Path(dir_path))

        if " rel " in mode:
            try:
                pl.Path(dir_path).parent.relative_to(pl.Path(self.project_path).parent)
            except ValueError:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    f"The directory <b>{pl.Path(dir_path).parent}</b> is not contained in <b>{pl.Path(self.project_path).parent}</b>.",
                )
                return

        if " rel " in mode:
            # convert to relative path (relative to BORIS project file)
            self.lw_images_directory.addItem(QListWidgetItem(str(pl.Path(dir_path).relative_to(pl.Path(self.project_path).parent))))
        else:
            self.lw_images_directory.addItem(QListWidgetItem(dir_path))
        self.lb_images_info.setText(f"Number of images in {dir_path}: {result.get('number of images', 0)}")

    def remove_images_directory(self):
        """
        remove dir path from the list
        """
        self.lw_images_directory.takeItem(self.lw_images_directory.currentRow())

    def add_button_menu(self, data, menu_obj):
        """
        add menu option from dictionary
        """
        if isinstance(data, dict):
            for k, v in data.items():
                sub_menu = QMenu(k, menu_obj)
                menu_obj.addMenu(sub_menu)
                self.add_button_menu(v, sub_menu)
        elif isinstance(data, list):
            for element in data:
                self.add_button_menu(element, menu_obj)
        else:
            action = menu_obj.addAction(data.split("|")[1])
            # tips are used to discriminate the menu option
            action.setStatusTip(data.split("|")[0])
            action.setIconVisibleInMenu(False)

    def cb_start_from_current_time_changed(self):
        """
        enable/disable radiobox for type of time selection
        """
        self.rb_day_time.setEnabled(self.cb_start_from_current_time.isChecked())
        self.rb_epoch_time.setEnabled(self.cb_start_from_current_time.isChecked())

    def limit_time_interval(self):
        """
        ask user a time interval for limiting the media observation
        """

        if self.cb_observation_time_interval.isChecked():
            time_interval_dialog = dialog.Ask_time(0)
            if self.time_format == cfg.S:
                time_interval_dialog.time_widget.rb_seconds.setChecked(True)
            if self.time_format == cfg.HHMMSS:
                time_interval_dialog.time_widget.rb_time.setChecked(True)
            time_interval_dialog.time_widget.set_time(0)
            time_interval_dialog.setWindowTitle("Start observation at")
            time_interval_dialog.label.setText("<b>Start</b> observation at")
            start_time, stop_time = 0, 0
            if time_interval_dialog.exec_():
                start_time = time_interval_dialog.time_widget.get_time()
            else:
                self.cb_observation_time_interval.setChecked(False)
                return
            time_interval_dialog.time_widget.set_time(0)
            time_interval_dialog.setWindowTitle("Stop observation at")
            time_interval_dialog.label.setText("<b>Stop</b> observation at")
            if time_interval_dialog.exec_():
                stop_time = time_interval_dialog.time_widget.get_time()
            else:
                self.cb_observation_time_interval.setChecked(False)
                return

            if start_time or stop_time:
                if stop_time <= start_time:
                    QMessageBox.critical(self, cfg.programName, "The stop time comes before the start time")
                    self.cb_observation_time_interval.setChecked(False)
                    return
                self.observation_time_interval = [start_time, stop_time]
                self.cb_observation_time_interval.setText(
                    (
                        "Limit observation to a time interval: "
                        f"{util.smart_time_format(start_time, self.time_format)} - {util.smart_time_format(stop_time, self.time_format)}"
                    )
                )
        else:
            self.observation_time_interval = [0, 0]
            self.cb_observation_time_interval.setText("Limit observation to a time interval")

    def tw_data_files_cellDoubleClicked(self, row, column):
        """
        double click on "Converters column"
        """
        if column == cfg.PLOT_DATA_CONVERTERS_IDX:
            if self.tw_data_files.item(row, cfg.PLOT_DATA_COLUMNS_IDX).text():
                w = AssignConverter(
                    self.tw_data_files.item(row, cfg.PLOT_DATA_COLUMNS_IDX).text(),
                    self.converters,
                    eval(self.tw_data_files.item(row, cfg.PLOT_DATA_CONVERTERS_IDX).text())
                    if self.tw_data_files.item(row, cfg.PLOT_DATA_CONVERTERS_IDX).text()
                    else "",
                )

                if w.exec_():
                    d = {}
                    for col_idx, cb in zip(self.tw_data_files.item(row, cfg.PLOT_DATA_COLUMNS_IDX).text().split(","), w.cbb):
                        if cb.currentText() != "None":
                            d[col_idx] = cb.currentText()
                    self.tw_data_files.item(row, cfg.PLOT_DATA_CONVERTERS_IDX).setText(str(d))
            else:
                QMessageBox.critical(self, cfg.programName, "Select the columns to plot (time,value)")

    def plot_data_file(self):
        """
        show plot
        check if data can be plotted
        """

        if self.pb_plot_data.text() != "Show plot":
            self.test.close_plot()
            self.text = None
            # update button text
            self.pb_plot_data.setText("Show plot")
            return

        if self.tw_data_files.selectedIndexes() or self.tw_data_files.rowCount() == 1:
            if self.tw_data_files.rowCount() == 1:
                row_idx = 0
            else:
                row_idx = self.tw_data_files.selectedIndexes()[0].row()

            filename = self.tw_data_files.item(row_idx, cfg.PLOT_DATA_FILEPATH_IDX).text()
            columns_to_plot = self.tw_data_files.item(row_idx, cfg.PLOT_DATA_COLUMNS_IDX).text()
            plot_title = self.tw_data_files.item(row_idx, cfg.PLOT_DATA_PLOTTITLE_IDX).text()

            # load converters in dictionary
            if self.tw_data_files.item(row_idx, cfg.PLOT_DATA_CONVERTERS_IDX).text():
                column_converter = eval(self.tw_data_files.item(row_idx, cfg.PLOT_DATA_CONVERTERS_IDX).text())
            else:
                column_converter = {}

            variable_name = self.tw_data_files.item(row_idx, cfg.PLOT_DATA_VARIABLENAME_IDX).text()
            time_interval = int(self.tw_data_files.item(row_idx, cfg.PLOT_DATA_TIMEINTERVAL_IDX).text())
            time_offset = int(self.tw_data_files.item(row_idx, cfg.PLOT_DATA_TIMEOFFSET_IDX).text())

            substract_first_value = self.tw_data_files.cellWidget(row_idx, cfg.PLOT_DATA_SUBSTRACT1STVALUE_IDX).currentText()

            plot_color = self.tw_data_files.cellWidget(row_idx, cfg.PLOT_DATA_PLOTCOLOR_IDX).currentText()

            data_file_path = project_functions.full_path(filename, self.project_path)

            if not data_file_path:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    (
                        f"Data file not found:\n{filename}\n"
                        "If the file path is not stored the data file "
                        "must be in the same directory than your project"
                    ),
                )
                return

            self.test = plot_data_module.Plot_data(
                data_file_path,
                time_interval,  # time interval
                time_offset,  # time offset
                plot_color,  # plot style
                plot_title,  # plot title
                variable_name,
                columns_to_plot,
                substract_first_value,
                self.converters,
                column_converter,
                log_level=logging.getLogger().getEffectiveLevel(),
            )

            if self.test.error_msg:
                QMessageBox.critical(self, cfg.programName, f"Impossible to plot data:\n{self.test.error_msg}")
                self.test = None
                return

            # self.test.setWindowFlags(self.test.windowFlags() | Qt.WindowStaysOnTopHint)
            self.test.show()
            self.test.update_plot(0)
            # update button text
            self.pb_plot_data.setText("Close plot")
        else:
            QMessageBox.warning(self, cfg.programName, "Select a data file")

    def add_data_file(self, mode: str):
        """
        user select a data file to be plotted synchronously with media file

        Args:
            mode (str): statusTip() data abs path / data rel path
        """

        if mode.split("|")[0] not in (
            "data abs path",
            "data rel path",
        ):
            QMessageBox.critical(
                self,
                cfg.programName,
                (f"Wrong mode to add a data file {mode}"),
            )
            return

        # check if project saved
        if (" w/o" in mode or " rel " in mode) and (not self.project_file_name):
            QMessageBox.critical(
                self,
                cfg.programName,
                ("It is not possible to add a data file with a relative path if the project is not already saved"),
            )
            return

        # limit to 2 files
        if self.tw_data_files.rowCount() >= 2:
            QMessageBox.warning(
                self,
                cfg.programName,
                ("It is not yet possible to plot more than 2 external data sources" "This limitation will be removed in future"),
            )
            return

        fd = QFileDialog()
        fd.setDirectory(os.path.expanduser("~") if (" abs " in mode) else str(pl.Path(self.project_path).parent))

        fn = fd.getOpenFileName(self, "Add data file", "", "All files (*)")
        file_name = fn[0] if type(fn) is tuple else fn

        if not file_name:
            return

        columns_to_plot = "1,2"  # columns to plot by default

        # check data file
        file_parameters = util.check_txt_file(file_name)

        if "error" in file_parameters:
            QMessageBox.critical(self, cfg.programName, f"Error on file {file_name}: {file_parameters['error']}")
            return

        if not file_parameters["homogeneous"]:  # the number of columns is not constant
            QMessageBox.critical(self, cfg.programName, "This file does not contain a constant number of columns")
            return

        header, footer = util.return_file_header_footer(file_name, file_row_number=file_parameters["rows number"], row_number=5)

        if not header:
            QMessageBox.critical(self, cfg.programName, f"Error on file {pl.Path(file_name).name}")
            return

        w = dialog.View_data()
        w.setWindowTitle("View data")
        w.lb.setText(f"View first and last rows of <b>{pl.Path(file_name).name}</b> file")

        w.tw.setColumnCount(file_parameters["fields number"])
        if footer:
            hf = header + [file_parameters["separator"].join(["..."] * file_parameters["fields number"])] + footer
            w.tw.setRowCount(len(header) + len(footer) + 1)
        else:
            hf = header
            w.tw.setRowCount(len(header))

        for idx, row in enumerate(hf):
            for col, v in enumerate(row.split(file_parameters["separator"])):
                item = QTableWidgetItem(v)
                item.setFlags(Qt.ItemIsEnabled)
                w.tw.setItem(idx, col, item)

        # stats
        try:
            df = pd.read_csv(file_name, sep=file_parameters["separator"], header=None if not file_parameters["has header"] else [0])
            # set columns names to based 1 index
            if not file_parameters["has header"]:
                df.columns = range(1, len(df.columns) + 1)

            stats_out = str(df.describe())
        except Exception:
            stats_out = "Not available"
        w.stats.setPlainText(stats_out)

        while True:
            flag_ok = True
            if w.exec_():
                columns_to_plot = w.le.text().replace(" ", "")
                for col in columns_to_plot.split(","):
                    try:
                        col_idx = int(col)
                    except ValueError:
                        QMessageBox.critical(self, cfg.programName, f"<b>{col}</b> does not seem to be a column index")
                        flag_ok = False
                        break
                    if col_idx <= 0 or col_idx > file_parameters["fields number"]:
                        QMessageBox.critical(self, cfg.programName, f"<b>{col}</b> is not a valid column index")
                        flag_ok = False
                        break
                if flag_ok:
                    break
            else:
                return

        else:
            return

        self.tw_data_files.setRowCount(self.tw_data_files.rowCount() + 1)

        if " rel " in mode:
            try:
                file_path = str(pl.Path(file_name).relative_to(pl.Path(self.project_path).parent))
            except ValueError:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    f"The directory <b>{pl.Path(file_name).parent}</b> is not contained in <b>{pl.Path(self.project_path).parent}</b>.",
                )
                return

        else:  # save absolute path
            file_path = file_name

        for col_idx, value in zip(
            [
                cfg.PLOT_DATA_FILEPATH_IDX,
                cfg.PLOT_DATA_COLUMNS_IDX,
                cfg.PLOT_DATA_PLOTTITLE_IDX,
                cfg.PLOT_DATA_VARIABLENAME_IDX,
                cfg.PLOT_DATA_CONVERTERS_IDX,
                cfg.PLOT_DATA_TIMEINTERVAL_IDX,
                cfg.PLOT_DATA_TIMEOFFSET_IDX,
            ],
            [file_path, columns_to_plot, "", "", "", "60", "0"],
        ):
            item = QTableWidgetItem(value)
            if col_idx == cfg.PLOT_DATA_CONVERTERS_IDX:
                item.setFlags(Qt.ItemIsEnabled)
                item.setBackground(QColor(230, 230, 230))
            self.tw_data_files.setItem(self.tw_data_files.rowCount() - 1, col_idx, item)

        # substract first value
        combobox = QComboBox()
        combobox.addItems(["True", "False"])
        self.tw_data_files.setCellWidget(self.tw_data_files.rowCount() - 1, cfg.PLOT_DATA_SUBSTRACT1STVALUE_IDX, combobox)

        # plot line color
        combobox = QComboBox()
        combobox.addItems(cfg.DATA_PLOT_STYLES)
        self.tw_data_files.setCellWidget(self.tw_data_files.rowCount() - 1, cfg.PLOT_DATA_PLOTCOLOR_IDX, combobox)

    def view_data_file_head_tail(self) -> None:
        """
        view first and last rows of data file
        """

        if not self.tw_data_files.selectedIndexes() and self.tw_data_files.rowCount() != 1:
            QMessageBox.warning(self, cfg.programName, "Select a data file")

        if self.tw_data_files.rowCount() == 1:
            data_file_path = project_functions.full_path(self.tw_data_files.item(0, 0).text(), self.project_path)
            columns_to_plot = self.tw_data_files.item(0, 1).text()
        else:  #  selected file
            data_file_path = project_functions.full_path(
                self.tw_data_files.item(self.tw_data_files.selectedIndexes()[0].row(), 0).text(), self.project_path
            )
            columns_to_plot = self.tw_data_files.item(self.tw_data_files.selectedIndexes()[0].row(), 1).text()

        file_parameters = util.check_txt_file(data_file_path)

        if "error" in file_parameters:
            QMessageBox.critical(self, cfg.programName, f"Error on file {data_file_path}: {file_parameters['error']}")
            return
        header, footer = util.return_file_header_footer(data_file_path, file_row_number=file_parameters["rows number"], row_number=5)

        if not header:
            QMessageBox.critical(self, cfg.programName, f"Error on file {pl.Path(data_file_path).name}")
            return

        w = dialog.View_data()
        w.setWindowTitle("View data")
        w.lb.setText(f"View first and last rows of <b>{pl.Path(data_file_path).name}</b> file")
        w.pbOK.setText("Close")
        w.label.setText("Index of columns to plot")
        w.le.setEnabled(False)
        w.le.setText(columns_to_plot)
        w.pbCancel.setVisible(False)

        w.tw.setColumnCount(file_parameters["fields number"])
        if footer:
            hf = header + [file_parameters["separator"].join(["..."] * file_parameters["fields number"])] + footer
            w.tw.setRowCount(len(header) + len(footer) + 1)
        else:
            hf = header
            w.tw.setRowCount(len(header))

        for idx, row in enumerate(hf):
            for col, v in enumerate(row.split(file_parameters["separator"])):
                item = QTableWidgetItem(v)
                item.setFlags(Qt.ItemIsEnabled)
                w.tw.setItem(idx, col, item)

        # stats
        try:
            df = pd.read_csv(
                data_file_path,
                sep=file_parameters["separator"],
                header=None if not file_parameters["has header"] else [0],
            )
            # set columns names to based 1 index
            if not file_parameters["has header"]:
                df.columns = range(1, len(df.columns) + 1)

            stats_out = str(df.describe())
        except Exception:
            stats_out = "Not available"
        w.stats.setPlainText(stats_out)

        w.exec_()

    def extract_wav(self):
        """
        extract wav of all media files loaded in player #1
        """

        if not self.cbVisualizeSpectrogram.isChecked() and not self.cb_visualize_waveform.isChecked():
            return

        flag_wav_produced = False
        # check if player 1 is selected
        flag_player1 = False
        for row in range(self.twVideo1.rowCount()):
            if self.twVideo1.cellWidget(row, 0).currentText() == "1":
                flag_player1 = True

        if not flag_player1:
            QMessageBox.critical(self, cfg.programName, "The player #1 is not selected")
            self.cbVisualizeSpectrogram.setChecked(False)
            self.cb_visualize_waveform.setChecked(False)
            return

        if True:
            w = dialog.Info_widget()
            w.resize(350, 100)
            # w.setWindowFlags(Qt.WindowStaysOnTopHint)
            w.setWindowTitle("BORIS")
            w.label.setText("Extracting WAV from media files...")

            for row in range(self.twVideo1.rowCount()):
                # check if player 1
                if self.twVideo1.cellWidget(row, 0).currentText() != "1":
                    continue

                media_file_path = project_functions.full_path(self.twVideo1.item(row, cfg.MEDIA_FILE_PATH_IDX).text(), self.project_path)
                if self.twVideo1.item(row, cfg.HAS_AUDIO_IDX).text() == "False":
                    QMessageBox.critical(self, cfg.programName, f"The media file {media_file_path} does not seem to have audio")
                    flag_wav_produced = False
                    break

                if os.path.isfile(media_file_path):
                    w.show()
                    QApplication.processEvents()

                    if util.extract_wav(self.ffmpeg_bin, media_file_path, self.tmp_dir) == "":
                        QMessageBox.critical(
                            self,
                            cfg.programName,
                            f"Error during extracting WAV of the media file {media_file_path}",
                        )
                        flag_wav_produced = False
                        break

                    w.hide()

                    flag_wav_produced = True
                else:
                    QMessageBox.warning(self, cfg.programName, f"<b>{media_file_path}</b> file not found")

            if not flag_wav_produced:
                self.cbVisualizeSpectrogram.setChecked(False)
                self.cb_visualize_waveform.setChecked(False)

    def check_creation_date(self) -> int:
        """
        check if media file exists
        check if Creation Date tag is present in metadata of media file

        Returns:
            int: 0 if OK else error code: 1 -> media file date not used, 2 -> media file not found

        """

        # check if media files exist

        media_not_found_list: list = []
        for row in range(self.twVideo1.rowCount()):
            if not pl.Path(self.twVideo1.item(row, 2).text()).is_file():
                media_not_found_list.append(self.twVideo1.item(row, 2).text())

        """
        if media_list:
            dlg = dialog.Results_dialog()
            dlg.setWindowTitle("BORIS")
            dlg.pbOK.setText("OK")
            dlg.pbCancel.setVisible(False)
            dlg.ptText.clear()
            dlg.ptText.appendHtml(
                (
                    "Some media file(s) were not found:<br>"
                    f"{'<br>'.join(media_list)}<br><br>"
                    "You cannot select the <b>Use the media creation date/time option</b>."
                )
            )
            dlg.ptText.moveCursor(QTextCursor.Start)
            ret = dlg.exec_()
        """

        not_tagged_media_list: list = []
        for row in range(self.twVideo1.rowCount()):
            if self.twVideo1.item(row, 2).text() not in media_not_found_list:
                media_info = util.accurate_media_analysis(self.ffmpeg_bin, self.twVideo1.item(row, 2).text())
                if cfg.MEDIA_CREATION_TIME not in media_info or media_info[cfg.MEDIA_CREATION_TIME] == cfg.NA:
                    not_tagged_media_list.append(self.twVideo1.item(row, 2).text())
                else:
                    creation_time_epoch = int(dt.datetime.strptime(media_info[cfg.MEDIA_CREATION_TIME], "%Y-%m-%d %H:%M:%S").timestamp())
                    self.media_creation_time[self.twVideo1.item(row, 2).text()] = creation_time_epoch

        if not_tagged_media_list:
            dlg = dialog.Results_dialog()
            dlg.setWindowTitle("BORIS")
            dlg.pbOK.setText("Yes")
            dlg.pbCancel.setVisible(True)
            dlg.pbCancel.setText("No")

            dlg.ptText.clear()
            dlg.ptText.appendHtml(
                (
                    "Some media file does not contain the <b>Creation date/time</b> metadata tag:<br>"
                    f"{'<br>'.join(not_tagged_media_list)}<br><br>"
                    "Use the media file date/time instead?"
                )
            )
            dlg.ptText.moveCursor(QTextCursor.Start)
            ret = dlg.exec_()

            if ret == 1:  #  use file creation time
                for media in not_tagged_media_list:
                    self.media_creation_time[media] = pl.Path(media).stat().st_ctime
                return 0  # OK use media file creation date/time
            else:
                self.cb_media_creation_date_as_offset.setChecked(False)
                self.media_creation_time = {}
                return 1
        else:
            return 0  # OK all media have a 'creation time' tag

    def closeEvent(self, event):
        """
        close observation windows
        """
        if self.test is not None:
            self.test.close_plot()
            self.text = None

    def pbCancel_clicked(self):
        """
        observation creation cancelled
        """
        if self.test is not None:
            self.test.close_plot()
            self.text = None
        self.reject()

    def check_parameters(self) -> bool:
        """
        check observation parameters

        Returns:
            bool: True if everything is OK else False

        """

        def is_numeric(s):
            """
            check if s is numeric (float)

            Args:
                s (str/int/float): value to test

            Returns:
                boolean: True if numeric else False
            """
            try:
                float(s)
                return True
            except ValueError:
                return False

        # check if observation id not empty
        if not self.leObservationId.text():
            self.qm = QMessageBox()
            self.qm.setIcon(QMessageBox.Critical)
            self.qm.setText("The <b>observation id</b> is mandatory and must be unique.")
            self.qm.exec_()
            return False

        # check if observation_type
        if not any((self.rb_media_files.isChecked(), self.rb_live.isChecked(), self.rb_images.isChecked())):
            self.qm = QMessageBox()
            self.qm.setIcon(QMessageBox.Critical)
            self.qm.setText("Choose an observation type.")
            self.qm.exec_()
            return False

        # check if offset is correct
        if self.cb_time_offset.isChecked():
            if self.obs_time_offset.get_time() is None:
                self.qm = QMessageBox()
                self.qm.setIcon(QMessageBox.Critical)
                self.qm.setText("Check the time offset value.")
                self.qm.exec_()
                return False

        if self.rb_media_files.isChecked():  # observation based on media file(s)
            # check if media file exists
            media_file_not_found: list = []
            for row in range(self.twVideo1.rowCount()):
                # check if media file exists
                if not pl.Path(self.twVideo1.item(row, 2).text()).is_file():
                    media_file_not_found.append(self.twVideo1.item(row, 2).text())

            # check player number
            players_list: list = []
            players: dict = {}  # for storing duration
            for row in range(self.twVideo1.rowCount()):
                player_idx = int(self.twVideo1.cellWidget(row, 0).currentText())
                players_list.append(player_idx)
                if player_idx not in players:
                    players[player_idx] = []
                players[player_idx].append(util.time2seconds(self.twVideo1.item(row, 3).text()))

            # check if player #1 is used
            if not players_list or min(players_list) > 1:
                self.qm = QMessageBox()
                self.qm.setIcon(QMessageBox.Critical)
                self.qm.setText("A media file must be loaded in player #1")
                self.qm.exec_()
                return False

            # check if players are used in crescent order
            if set(list(range(min(players_list), max(players_list) + 1))) != set(players_list):
                self.qm = QMessageBox()
                self.qm.setIcon(QMessageBox.Critical)
                self.qm.setText("Some player are not used. Please reorganize your media files")
                self.qm.exec_()
                return False

            # check if more media in player #1 and media in other players
            """
            if len(players[1]) > 1 and set(players.keys()) != {1}:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    (
                        "It is not possible to play another media synchronously "
                        "when many media are queued in the first media player"
                    ),
                )
                return False
            """

            # check if more media enqueued on many players
            if len(set(players.keys())) > 1:  # many players used
                if max([len(players[x]) for x in players]) > 1:
                    QMessageBox.critical(
                        self,
                        cfg.programName,
                        (
                            "It is not possible to enqueue media files "
                            "on more than one player.<br>"
                            "You can use the <b>Merge media files</b> tool (see Tools > Media file > Merge media files"
                        ),
                    )
                    return False

            # check that the longuest media is in player #1
            durations: list = []
            for i in sorted(list(players.keys())):
                durations.append(sum(players[i]))
            if [x for x in durations[1:] if x > durations[0]]:
                QMessageBox.critical(self, cfg.programName, "The longuest media file(s) must be loaded in player #1")
                return False

            # check offset for media files
            for row in range(self.twVideo1.rowCount()):
                if not is_numeric(self.twVideo1.item(row, 1).text()):
                    QMessageBox.critical(
                        self,
                        cfg.programName,
                        (
                            "The offset value "
                            f"<b>{self.twVideo1.item(row, 1).text()}</b>"
                            " is not recognized as a numeric value.<br>"
                            "Use decimal number of seconds (e.g. -58.5 or 32)"
                        ),
                    )
                    return False

            # check if offset set and only player #1 is used
            if len(set(players_list)) == 1:
                for row in range(self.twVideo1.rowCount()):
                    if float(self.twVideo1.item(row, 1).text()):
                        QMessageBox.critical(
                            self,
                            cfg.programName,
                            (
                                "It is not possible to use offset value(s) with only one player,<br>"
                                "The offset values are use to synchronise various players."
                            ),
                        )
                        return False

            # check offset for external data files
            for row in range(self.tw_data_files.rowCount()):
                if not is_numeric(self.tw_data_files.item(row, cfg.PLOT_DATA_TIMEOFFSET_IDX).text()):
                    QMessageBox.critical(
                        self,
                        cfg.programName,
                        (
                            "The external data file start value "
                            f"<b>{self.tw_data_files.item(row, cfg.PLOT_DATA_TIMEOFFSET_IDX).text()}</b>"
                            " is not recognized as a numeric value.<br>"
                            "Use decimal number of seconds (e.g. -58.5 or 32)"
                        ),
                    )
                    return False

            # check media creation time tag in metadata
            # Disable because the check will be made at the oservation start
            """
            if self.cb_media_creation_date_as_offset.isChecked():
                if self.check_creation_date():
                    return False
            """

        if self.rb_images.isChecked():  # observation based on images directory
            if not self.lw_images_directory.count():
                QMessageBox.critical(self, cfg.programName, "You have to select at least one images directory")
                return False

        # check if indep variables are correct type
        for row in range(self.twIndepVariables.rowCount()):
            if self.twIndepVariables.item(row, 1).text() == cfg.NUMERIC:
                if self.twIndepVariables.item(row, 2).text() and not is_numeric(self.twIndepVariables.item(row, 2).text()):
                    QMessageBox.critical(
                        self,
                        cfg.programName,
                        f"The <b>{self.twIndepVariables.item(row, 0).text()}</b> variable must be numeric!",
                    )
                    return False

        # check if new obs and observation id already present or if edit obs and id changed
        if (self.mode == "new") or (self.mode == "edit" and self.leObservationId.text() != self.mem_obs_id):
            if self.leObservationId.text() in self.pj[cfg.OBSERVATIONS]:
                QMessageBox.critical(
                    self,
                    cfg.programName,
                    (
                        f"The observation id <b>{self.leObservationId.text()}</b> is already used!<br>"
                        f"{self.pj[cfg.OBSERVATIONS][self.leObservationId.text()]['description']}<br>"
                        f"{self.pj[cfg.OBSERVATIONS][self.leObservationId.text()]['date']}"
                    ),
                )
                return False

        # check if numeric indep variable values are numeric
        for row in range(self.twIndepVariables.rowCount()):
            if self.twIndepVariables.item(row, 1).text() == cfg.NUMERIC:
                if self.twIndepVariables.item(row, 2).text() and not is_numeric(self.twIndepVariables.item(row, 2).text()):
                    QMessageBox.critical(
                        self,
                        cfg.programName,
                        f"The <b>{self.twIndepVariables.item(row, 0).text()}</b> variable must be numeric!",
                    )
                    return False

        return True

    def pbLaunch_clicked(self):
        """
        Close dialog and start the observation
        """

        if self.check_parameters():
            if self.test is not None:
                self.test.close_plot()
                self.text = None
            self.done(2)

    def pbSave_clicked(self):
        """
        Close window and save observation
        """
        if self.check_parameters():
            self.state = "accepted"
            if self.test is not None:
                self.test.close_plot()
                self.text = None
            self.accept()
        else:
            self.state = "refused"

    def check_media(self, file_path: str, mode: str) -> tuple:
        """
        check media and add them to list view if duration > 0

        Args:
            file_path (str): media file path to be checked
            mode (str): mode for adding media file

        Returns:
             bool: False if file is media else True
             str: error message or empty string
        """

        media_info = util.accurate_media_analysis(self.ffmpeg_bin, file_path)

        if "error" in media_info:
            return (False, media_info["error"])

        if media_info["format_long_name"] == "Tele-typewriter":
            return (False, "Text file")

        if media_info["duration"] > 0:
            if " rel " in mode:
                # convert to relative path (relative to BORIS project file)
                file_path = str(pl.Path(file_path).relative_to(pl.Path(self.project_path).parent))

            self.mediaDurations[file_path] = float(media_info["duration"])
        elif media_info["has_video"] is False and media_info["audio_duration"]:
            self.mediaDurations[file_path] = float(media_info["audio_duration"])
        else:
            return (True, "Media duration not available")

        self.mediaFPS[file_path] = float(media_info["fps"])
        self.mediaHasVideo[file_path] = media_info["has_video"]
        self.mediaHasAudio[file_path] = media_info["has_audio"]
        self.add_media_to_listview(file_path)
        return (False, "")

    def update_media_options(self):
        """
        update the media options
        """
        for w in (
            self.cbVisualizeSpectrogram,
            self.cb_visualize_waveform,
            self.cb_observation_time_interval,
            self.cb_media_creation_date_as_offset,
        ):
            w.setEnabled(self.twVideo1.rowCount() > 0)

        # enable stop ongoing state events if n. media > 1
        self.cbCloseCurrentBehaviorsBetweenVideo.setEnabled(self.twVideo1.rowCount() > 0)

        # DEVELOPMENT (REMOVE BEFORE RELEASE)
        self.cb_media_creation_date_as_offset.setEnabled(False)

    def add_media(self, mode: str):
        """
        add media

        Args:
            mode (str): mode for adding the media file
        """

        if mode.split("|")[0] not in (
            "media abs path",
            "media rel path",
            "dir abs path",
            "dir rel path",
        ):
            QMessageBox.critical(
                self,
                cfg.programName,
                (f"Wrong mode to add media {mode}"),
            )
            return

        # check if project saved
        if (" w/o" in mode or " rel " in mode) and (not self.project_file_name):
            QMessageBox.critical(
                self,
                cfg.programName,
                ("It is not possible to add a media file without path or with a relative path if the project is not already saved"),
            )
            return

        fd = QFileDialog()
        if self.mem_dir:
            fd.setDirectory(self.mem_dir if (" abs " in mode) else str(pl.Path(self.project_path).parent))
        else:
            fd.setDirectory(os.path.expanduser("~") if (" abs " in mode) else str(pl.Path(self.project_path).parent))

        if "media " in mode:
            fn = fd.getOpenFileNames(self, "Add media file(s)", "", "All files (*)")
            file_paths = fn[0] if type(fn) is tuple else fn

            if file_paths:
                # store directory for next usage
                self.mem_dir = str(pl.Path(file_paths[0]).parent)
                # check if media dir in contained in the BORIS file project dir
                if " rel " in mode:
                    try:
                        pl.Path(file_paths[0]).parent.relative_to(pl.Path(self.project_path).parent)
                    except ValueError:
                        QMessageBox.critical(
                            self,
                            cfg.programName,
                            f"The directory <b>{pl.Path(file_paths[0]).parent}</b> is not contained in <b>{pl.Path(self.project_path).parent}</b>.",
                        )
                        return

                for file_path in file_paths:
                    (error, msg) = self.check_media(file_path, mode)
                    if error:
                        QMessageBox.critical(self, cfg.programName, f"<b>{file_path}</b>. {msg}")

        if "dir " in mode:  # add media from dir
            dir_name = fd.getExistingDirectory(self, "Select directory")
            if dir_name:
                response = ""
                for file_path in sorted(pl.Path(dir_name).glob("*")):
                    if not file_path.is_file():
                        continue
                    (error, msg) = self.check_media(str(file_path), mode)
                    if error:
                        if response != "Skip all non media files":
                            response = dialog.MessageDialog(
                                cfg.programName,
                                f"<b>{file_path}</b> {msg}",
                                ["Continue", "Skip all non media files", cfg.CANCEL],
                            )
                            if response == cfg.CANCEL:
                                break

        self.update_media_options()

    def add_media_to_listview(self, file_name):
        """
        add media file path to list widget
        """
        # add a row
        self.twVideo1.setRowCount(self.twVideo1.rowCount() + 1)

        for col_idx, s in enumerate(
            (
                None,
                0,
                file_name,
                util.seconds2time(self.mediaDurations[file_name]),
                f"{self.mediaFPS[file_name]:.2f}",
                self.mediaHasVideo[file_name],
                self.mediaHasAudio[file_name],
            )
        ):
            if col_idx == 0:  # player combobox
                combobox = QComboBox()
                combobox.addItems(cfg.ALL_PLAYERS)
                self.twVideo1.setCellWidget(self.twVideo1.rowCount() - 1, col_idx, combobox)
            else:
                item = QTableWidgetItem(f"{s}")
                if col_idx != 1:  # only offset is editable by user
                    item.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable)

                self.twVideo1.setItem(self.twVideo1.rowCount() - 1, col_idx, item)

    def remove_data_file(self):
        """
        remove all selected data file from list widget
        """
        if self.tw_data_files.selectedIndexes():
            rows_to_delete = set([x.row() for x in self.tw_data_files.selectedIndexes()])
            for row in sorted(rows_to_delete, reverse=True):
                self.tw_data_files.removeRow(row)
        else:
            QMessageBox.warning(self, cfg.programName, "No data file selected")

    def remove_media(self):
        """
        remove all selected media files from list widget
        """

        if not self.twVideo1.selectedIndexes():
            QMessageBox.warning(self, cfg.programName, "No media file selected")
            return

        rows_to_delete = set([x.row() for x in self.twVideo1.selectedIndexes()])
        for row in sorted(rows_to_delete, reverse=True):
            media_path = self.twVideo1.item(row, cfg.MEDIA_FILE_PATH_IDX).text()
            self.twVideo1.removeRow(row)
            if media_path not in [self.twVideo1.item(idx, cfg.MEDIA_FILE_PATH_IDX).text() for idx in range(self.twVideo1.rowCount())]:
                try:
                    del self.mediaDurations[media_path]
                except NameError:
                    pass
                try:
                    del self.mediaFPS[media_path]
                except NameError:
                    pass

        self.update_media_options()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import io
import pandas as pd
import pathlib as pl

try:
    import pyreadr

    flag_pyreadr_loaded = True
except ModuleNotFoundError:
    flag_pyreadr_loaded = False


from PyQt5.QtCore import QPoint, Qt, pyqtSignal, QEvent
from PyQt5.QtGui import QColor, QPainter, QPolygon, QPixmap
from PyQt5.QtWidgets import (
    QApplication,
    QCheckBox,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMessageBox,
    QTableWidget,
    QTableWidgetItem,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
    QColorDialog,
    QSpacerItem,
    QSizePolicy,
    QAction,
    QDialog,
)

from typing import List

from . import config as cfg
from . import dialog, menu_options
from . import utilities as util


class wgMeasurement(QDialog):
    """
    widget for geometric measurements
    """

    closeSignal = pyqtSignal()
    send_event_signal = pyqtSignal(QEvent)
    reload_image_signal = pyqtSignal()
    save_picture_signal = pyqtSignal(str)
    mark_color: str = cfg.ACTIVE_MEASUREMENTS_COLOR
    flag_saved = True  # store if measurements are saved
    draw_mem: dict = {}
    mem_points: list = []  # memory of clicked points
    mem_video: list = []  # memory of clicked points

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Geometric measurements")

        vbox = QVBoxLayout(self)

        self.rb_point = QRadioButton("Point (left click)", clicked=self.rb_clicked)
        vbox.addWidget(self.rb_point)

        self.rb_polyline = QRadioButton("Polyline (left click for vertices, right click to finish)", clicked=self.rb_clicked)
        vbox.addWidget(self.rb_polyline)

        self.rb_polygon = QRadioButton(
            "Polygon (left click for Polygon vertices, right click to close the polygon)", clicked=self.rb_clicked
        )
        vbox.addWidget(self.rb_polygon)

        self.rb_angle = QRadioButton("Angle (vertex: left click, segments: right click)", clicked=self.rb_clicked)
        vbox.addWidget(self.rb_angle)

        self.rb_oriented_angle = QRadioButton("Oriented angle (vertex: left click, segments: right click)", clicked=self.rb_clicked)
        vbox.addWidget(self.rb_oriented_angle)

        hbox = QHBoxLayout()
        self.cbPersistentMeasurements = QCheckBox("Measurements are persistent")
        self.cbPersistentMeasurements.setChecked(True)
        hbox.addWidget(self.cbPersistentMeasurements)

        # color chooser
        self.bt_color_chooser = QPushButton("Choose color of marks", clicked=self.choose_marks_color)
        self.bt_color_chooser.setStyleSheet(f"QWidget {{background-color:{self.mark_color}}}")
        hbox.addWidget(self.bt_color_chooser)

        hbox.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        vbox.addLayout(hbox)

        vbox.addWidget(QLabel("<b>Scale</b>"))

        hbox1 = QHBoxLayout()
        self.lbRef = QLabel("Reference")
        hbox1.addWidget(self.lbRef)
        self.lbPx = QLabel("Pixels")
        hbox1.addWidget(self.lbPx)
        vbox.addLayout(hbox1)

        hbox2 = QHBoxLayout()
        self.leRef = QLineEdit()
        self.leRef.setText("1")
        hbox2.addWidget(self.leRef)
        self.lePx = QLineEdit()
        self.lePx.setText("1")
        hbox2.addWidget(self.lePx)
        vbox.addLayout(hbox2)

        self.pte = QTableWidget()
        self.pte.verticalHeader().hide()

        # header
        self.measurements_header = [
            "Player",
            "media file name",
            "Time",
            "Frame index",
            "Type of measurement",
            "x",
            "y",
            "Distance",
            "Area",
            "Angle",
            "Coordinates",
        ]
        self.pte.setColumnCount(len(self.measurements_header))
        self.pte.setHorizontalHeaderLabels(self.measurements_header)

        self.pte.setSelectionBehavior(QTableWidget.SelectRows)
        self.pte.setSelectionMode(QTableWidget.MultiSelection)

        self.pte.setContextMenuPolicy(Qt.ActionsContextMenu)

        self.action = QAction("Delete measurement")
        self.action.triggered.connect(self.delete_measurement)

        self.pte.addAction(self.action)

        vbox.addWidget(self.pte)

        self.status_lb = QLabel()
        vbox.addWidget(self.status_lb)

        hbox3 = QHBoxLayout()
        hbox3.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        self.pb_clear = QPushButton("Clear measurements", clicked=self.pbClear_clicked)
        hbox3.addWidget(self.pb_clear)

        self.pb_save_picture = QPushButton("Save current picture", clicked=self.pb_save_picture_clicked)
        hbox3.addWidget(self.pb_save_picture)

        # disabled for now
        self.pb_save_all_pictures = QPushButton("Save all pictures", clicked=self.pb_save_all_pictures_clicked)
        hbox3.addWidget(self.pb_save_all_pictures)

        self.pb_save = QPushButton("Save results", clicked=self.pb_save_clicked)
        hbox3.addWidget(self.pb_save)
        self.pb_close = QPushButton("Close", clicked=self.pbClose_clicked)
        hbox3.addWidget(self.pb_close)
        vbox.addLayout(hbox3)

        self.installEventFilter(self)

    def rb_clicked(self):
        """
        radiobutton clicked, all points in memory are cleared
        """
        self.mem_points = []
        self.mem_video = []

        self.reload_image_signal.emit()

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.send_event_signal.emit(event)
            return True
        else:
            return False

    def pb_save_picture_clicked(self):
        """
        ask to save the current frame
        """
        self.save_picture_signal.emit("current")

    def pb_save_all_pictures_clicked(self):
        """
        ask to save all frames with measurements
        """
        self.save_picture_signal.emit("all")

    def delete_measurement(self):
        """
        delete the selected measurement(s)
        """

        if not self.pte.selectedItems():
            return

        rows_to_delete: list = []
        for item in self.pte.selectedItems():
            if item.row() not in rows_to_delete:
                rows_to_delete.append(item.row())

        elements_to_delete = []
        for row in sorted(rows_to_delete, reverse=True):
            player = int(self.pte.item(row, 0).text())
            frame_idx = int(self.pte.item(row, 2).text())
            obj_type = self.pte.item(row, 3).text()
            coord = eval(self.pte.item(row, 9).text())

            if frame_idx in self.draw_mem:
                for idx, element in enumerate(self.draw_mem[frame_idx]):
                    if (element["player"] == player - 1) and (element["object_type"] == obj_type) and (element["coordinates"] == coord):
                        elements_to_delete.append((frame_idx, idx))

            self.pte.removeRow(row)
            self.pte.flag_saved = False

        for frame_idx, idx in sorted(elements_to_delete, reverse=True):
            self.draw_mem[frame_idx].pop(idx)

        self.reload_image_signal.emit()

    def choose_marks_color(self):
        """
        show the color chooser dialog
        """
        cd = QColorDialog()
        cd.setWindowFlags(Qt.WindowStaysOnTopHint)
        cd.setOptions(QColorDialog.ShowAlphaChannel | QColorDialog.DontUseNativeDialog)
        cd.setCurrentColor(QColor(self.mark_color))

        if cd.exec_():
            new_color = cd.currentColor()
            self.bt_color_chooser.setStyleSheet(f"QWidget {{background-color:{new_color.name()}}}")
            self.mark_color = new_color.name()

    def closeEvent(self, event):
        """
        Intercept the close event to check if measurements are saved
        """

        logging.debug("close event")

        if not self.flag_saved:
            response = dialog.MessageDialog(
                cfg.programName,
                "The current measurements are not saved. Do you want to save the measurement results before closing?",
                (cfg.YES, cfg.NO, cfg.CANCEL),
            )
            if response == cfg.YES:
                if self.pb_save_clicked():
                    event.ignore()
                    return
            if response == cfg.CANCEL:
                event.ignore()
                return

        self.flag_saved: bool = True
        self.draw_mem: dict = {}
        self.closeSignal.emit()

    def pbClear_clicked(self):
        """
        clear measurements draw and results
        """

        if not self.flag_saved:
            response = dialog.MessageDialog(
                cfg.programName,
                "Confirm clearing",
                (cfg.YES, cfg.CANCEL),
            )
            if response == cfg.CANCEL:
                return

        self.draw_mem: dict = {}
        self.mem_points: list = []
        self.mem_video: list = []

        self.pte.clear()
        self.pte.setColumnCount(len(self.measurements_header))
        self.pte.setRowCount(0)
        self.pte.setHorizontalHeaderLabels(self.measurements_header)
        self.flag_saved = True

        self.reload_image_signal.emit()

    def pbClose_clicked(self):
        """
        Close button
        """
        logging.debug("close function")
        self.close()

    def pb_save_clicked(self) -> bool:
        """
        Save measurements results
        """

        file_formats = [cfg.TSV, cfg.CSV, cfg.ODS, cfg.XLSX, cfg.HTML, cfg.PANDAS_DF]
        if flag_pyreadr_loaded:
            file_formats.append(cfg.RDS)

        # default file name
        media_file_list: list = []
        for row in range(self.pte.rowCount()):
            media_file_list.append(self.pte.item(row, 1).text())

        if len(set(media_file_list)) == 1:
            default_file_name = str(pl.Path(media_file_list[0]).with_suffix(""))
        else:
            default_file_name = ""

        file_name, filter_ = QFileDialog().getSaveFileName(self, "Save geometric measurements", default_file_name, ";;".join(file_formats))
        if not file_name:
            return True

        # add correct file extension if not present
        if pl.Path(file_name).suffix != f".{cfg.FILE_NAME_SUFFIX[filter_]}":
            file_name = str(pl.Path(file_name)) + "." + cfg.FILE_NAME_SUFFIX[filter_]
            # check if file with new extension already exists
            if pl.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", (cfg.CANCEL, cfg.OVERWRITE))
                    == cfg.CANCEL
                ):
                    return True

        plain_text: str = "\t".join(self.measurements_header) + "\n"
        for row in range(self.pte.rowCount()):
            row_content: list = []
            for col in range(self.pte.columnCount()):
                row_content.append(self.pte.item(row, col).text())
            plain_text += "\t".join(row_content) + "\n"

        plain_text = plain_text[:-1]

        df = pd.read_csv(io.StringIO(plain_text), sep="\t")

        try:
            if filter_ == cfg.ODS:
                df.to_excel(file_name, engine="odf", sheet_name="Geometric measurements", index=False, na_rep="NA")
                self.flag_saved = True
            if filter_ == cfg.XLSX:
                df.to_excel(file_name, sheet_name="Geometric measurements", index=False, na_rep="NA")
                self.flag_saved = True
            if filter_ == cfg.HTML:
                df.to_html(file_name, index=False, na_rep="NA")
                self.flag_saved = True
            if filter_ == cfg.CSV:
                df.to_csv(file_name, index=False, sep=",", na_rep="NA")
                self.flag_saved = True
            if filter_ == cfg.TSV:
                df.to_csv(file_name, index=False, sep="\t", na_rep="NA")
                self.flag_saved = True
            if filter_ == cfg.PANDAS_DF:
                df.to_pickle(file_name)
            if filter_ == cfg.RDS:
                pyreadr.write_rds(file_name, df)

        except Exception:
            QMessageBox.warning(self, cfg.programName, "An error occured during saving the measurement results")
            return True
        return False  # everything OK


def show_widget(self) -> None:
    """
    active the geometric measurement widget
    """

    def close_measurement_widget():
        """
        close the geometric measurement widget
        """

        logging.debug("close_measurement_widget")

        if self.observationId and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
            for n_player, dw in enumerate(self.dw_player):
                dw.frame_viewer.clear()
                dw.stack.setCurrentIndex(cfg.VIDEO_VIEWER)
                dw.setWindowTitle(f"Player #{n_player + 1}")
            self.actionPlay.setEnabled(True)

        if self.observationId and self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
            for dw in self.dw_player:
                self.extract_frame(dw)

        self.geometric_measurements_mode = False
        self.measurement_w.draw_mem = {}

        self.measurement_w.close()
        menu_options.update_menu(self)

    self.geometric_measurements_mode = True
    self.pause_video()

    menu_options.update_menu(self)

    self.actionPlay.setEnabled(False)

    self.measurement_w = wgMeasurement()
    # self.measurement_w.setWindowFlags(Qt.WindowStaysOnTopHint)
    self.measurement_w.closeSignal.connect(close_measurement_widget)
    self.measurement_w.send_event_signal.connect(self.signal_from_widget)
    self.measurement_w.reload_image_signal.connect(self.reload_frame)
    self.measurement_w.save_picture_signal.connect(self.save_picture_with_measurements)
    self.measurement_w.draw_mem = {}

    self.measurement_w.show()

    for dw in self.dw_player:
        dw.setWindowTitle("Geometric measurements")
        dw.stack.setCurrentIndex(cfg.PICTURE_VIEWER)
        self.extract_frame(dw)


def draw_point(self, x: int, y: int, color: str, n_player: int = 0) -> None:
    """
    draw point on frame-by-frame image
    """
    RADIUS = 6
    painter = QPainter()
    painter.begin(self.dw_player[n_player].frame_viewer.pixmap())
    painter.setPen(QColor(color))
    painter.drawEllipse(QPoint(x, y), RADIUS, RADIUS)
    # cross inside circle
    painter.drawLine(x - RADIUS, y, x + RADIUS, y)
    painter.drawLine(x, y - RADIUS, x, y + RADIUS)
    painter.end()
    self.dw_player[n_player].frame_viewer.update()


def draw_line(self, x1: int, y1: int, x2: int, y2: int, color: str, n_player: int = 0) -> None:
    """
    draw line on frame-by-frame image
    """
    painter = QPainter()
    painter.begin(self.dw_player[n_player].frame_viewer.pixmap())
    painter.setPen(QColor(color))
    painter.drawLine(x1, y1, x2, y2)
    painter.end()
    self.dw_player[n_player].frame_viewer.update()


def append_results(self, results: list) -> None:
    """
    append results to measurements table
    """
    self.measurement_w.pte.setRowCount(self.measurement_w.pte.rowCount() + 1)
    for idx, x in enumerate(results):
        item = QTableWidgetItem()
        item.setText(str(x))
        item.setFlags(Qt.ItemIsSelectable | Qt.ItemIsEnabled)
        self.measurement_w.pte.setItem(self.measurement_w.pte.rowCount() - 1, idx, item)


def image_clicked(self, n_player: int, event) -> None:
    """
    Geometric measurements on image

    Args:
        n_player (int): id of clicked player
        event (Qevent): event (mousepressed)
    """

    logging.debug("function image_clicked")

    if not self.geometric_measurements_mode:
        return

    if self.mem_player != -1 and n_player != self.mem_player:
        self.mem_player = n_player
        return

    self.mem_player = n_player
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
        if self.dw_player[n_player].player.estimated_frame_number is not None:
            current_frame = self.dw_player[n_player].player.estimated_frame_number
        else:
            current_frame = cfg.NA
    elif self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
        current_frame = self.image_idx

    if not (hasattr(self, "measurement_w") and (self.measurement_w is not None) and (self.measurement_w.isVisible())):
        return

    x, y = event.pos().x(), event.pos().y()

    # convert label coordinates in pixmap coordinates
    pixmap_x = int(x - (self.dw_player[n_player].frame_viewer.width() - self.dw_player[n_player].frame_viewer.pixmap().width()) / 2)
    pixmap_y = int(y - (self.dw_player[n_player].frame_viewer.height() - self.dw_player[n_player].frame_viewer.pixmap().height()) / 2)

    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
        # convert pixmap coordinates in video coordinates
        x_video = round((pixmap_x / self.dw_player[n_player].frame_viewer.pixmap().width()) * self.dw_player[n_player].player.width)
        y_video = round((pixmap_y / self.dw_player[n_player].frame_viewer.pixmap().height()) * self.dw_player[n_player].player.height)
    elif self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
        original_width = QPixmap(self.images_list[self.image_idx]).size().width()
        original_height = QPixmap(self.images_list[self.image_idx]).size().height()
        x_video = round((pixmap_x / self.dw_player[n_player].frame_viewer.pixmap().width()) * original_width)
        y_video = round((pixmap_y / self.dw_player[n_player].frame_viewer.pixmap().height()) * original_height)

    if not (
        0 <= pixmap_x <= self.dw_player[n_player].frame_viewer.pixmap().width()
        and 0 <= pixmap_y <= self.dw_player[n_player].frame_viewer.pixmap().height()
    ):
        self.measurement_w.status_lb.setText("<b>The click is outside the video area</b>")
        return

    self.measurement_w.status_lb.clear()

    # media file name
    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
        fn = self.dw_player[n_player - 1].player.playlist[self.dw_player[n_player - 1].player.playlist_pos]["filename"]

        print(f"{fn=}")

        # check if media file path contained in media list
        if [True for x in self.pj[cfg.OBSERVATIONS][self.observationId]["file"].values() if fn in x]:
            media_file_name = self.dw_player[n_player - 1].player.playlist[self.dw_player[n_player - 1].player.playlist_pos]["filename"]
        else:
            media_file_name = str(pl.Path(fn).relative_to(pl.Path(self.projectFileName).parent))

            print(f"{media_file_name=}")

    if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
        # check if pictures dir path is relative
        if str(pl.Path(self.images_list[current_frame]).parent) not in self.pj[cfg.OBSERVATIONS][self.observationId].get(
            cfg.DIRECTORIES_LIST, []
        ):
            media_file_name = str(pl.Path(self.images_list[current_frame]).relative_to(pl.Path(self.projectFileName).parent))
        else:
            media_file_name = self.images_list[current_frame]

    # point
    if self.measurement_w.rb_point.isChecked():
        if event.button() == Qt.LeftButton:
            draw_point(self, pixmap_x, pixmap_y, self.measurement_w.mark_color, n_player)
            if current_frame not in self.measurement_w.draw_mem:
                self.measurement_w.draw_mem[current_frame] = []

            self.measurement_w.draw_mem[current_frame].append(
                {
                    "player": n_player,
                    "object_type": cfg.POINT_OBJECT,
                    "color": self.measurement_w.mark_color,
                    "coordinates": [(x_video, y_video)],
                }
            )

            append_results(
                self,
                (
                    n_player + 1,
                    media_file_name,
                    f"{self.getLaps():.03f}",
                    current_frame,
                    cfg.POINT_OBJECT,
                    x_video,
                    y_video,
                    cfg.NA,
                    cfg.NA,
                    cfg.NA,
                    str([(x_video, y_video)]),
                ),
            )

            self.measurement_w.flag_saved = False

    # angle
    elif self.measurement_w.rb_angle.isChecked() or self.measurement_w.rb_oriented_angle.isChecked():
        if event.button() == Qt.LeftButton:
            draw_point(self, pixmap_x, pixmap_y, self.measurement_w.mark_color, n_player)
            self.measurement_w.mem_points = [(pixmap_x, pixmap_y)]
            self.measurement_w.mem_video = [(x_video, y_video)]

        if event.button() == Qt.RightButton and len(self.measurement_w.mem_points):
            draw_point(self, pixmap_x, pixmap_y, self.measurement_w.mark_color, n_player)
            draw_line(
                self,
                self.measurement_w.mem_points[0][0],
                self.measurement_w.mem_points[0][1],
                pixmap_x,
                pixmap_y,
                self.measurement_w.mark_color,
                n_player,
            )

            self.measurement_w.mem_points.append((pixmap_x, pixmap_y))
            self.measurement_w.mem_video.append((x_video, y_video))

            if len(self.measurement_w.mem_points) == 3:
                if self.measurement_w.rb_angle.isChecked():
                    angle = util.angle(self.measurement_w.mem_points[0], self.measurement_w.mem_points[1], self.measurement_w.mem_points[2])
                    object_type = cfg.ANGLE_OBJECT
                else:  # oriented angle
                    angle = util.oriented_angle(
                        self.measurement_w.mem_points[0], self.measurement_w.mem_points[1], self.measurement_w.mem_points[2]
                    )
                    object_type = cfg.ORIENTED_ANGLE_OBJECT

                append_results(
                    self,
                    (
                        n_player + 1,
                        media_file_name,
                        f"{self.getLaps():.03f}",
                        current_frame,
                        object_type,
                        cfg.NA,
                        cfg.NA,
                        cfg.NA,
                        cfg.NA,
                        round(
                            angle,
                            1,
                        ),
                        str(self.measurement_w.mem_video),
                    ),
                )

                self.measurement_w.flag_saved = False
                if current_frame not in self.measurement_w.draw_mem:
                    self.measurement_w.draw_mem[current_frame] = []

                self.measurement_w.draw_mem[current_frame].append(
                    {
                        "player": n_player,
                        "object_type": object_type,
                        "color": self.measurement_w.mark_color,
                        "coordinates": self.measurement_w.mem_video,
                    }
                )

                self.measurement_w.mem_points, self.measurement_w.mem_video = [], []

    # polygon
    elif self.measurement_w.rb_polygon.isChecked():
        if event.button() == Qt.LeftButton:
            draw_point(self, pixmap_x, pixmap_y, self.measurement_w.mark_color, n_player)
            if len(self.measurement_w.mem_points):
                draw_line(
                    self,
                    self.measurement_w.mem_points[-1][0],
                    self.measurement_w.mem_points[-1][1],
                    pixmap_x,
                    pixmap_y,
                    self.measurement_w.mark_color,
                    n_player,
                )
            self.measurement_w.mem_points.append((pixmap_x, pixmap_y))
            self.measurement_w.mem_video.append((x_video, y_video))

        if event.button() == Qt.RightButton and len(self.measurement_w.mem_points) >= 2:
            # close polygon
            draw_line(
                self,
                self.measurement_w.mem_points[-1][0],
                self.measurement_w.mem_points[-1][1],
                self.measurement_w.mem_points[0][0],
                self.measurement_w.mem_points[0][1],
                self.measurement_w.mark_color,
                n_player,
            )
            if current_frame not in self.measurement_w.draw_mem:
                self.measurement_w.draw_mem[current_frame] = []

            self.measurement_w.draw_mem[current_frame].append(
                {
                    "player": n_player,
                    "object_type": cfg.POLYGON_OBJECT,
                    "color": self.measurement_w.mark_color,
                    "coordinates": self.measurement_w.mem_video,
                }
            )

            area = util.polygon_area(self.measurement_w.mem_video)
            try:
                area = area / (float(self.measurement_w.lePx.text()) ** 2) * float(self.measurement_w.leRef.text()) ** 2
            except Exception:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    "Check reference and pixel values! Values must be numeric.",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )

            length = util.polyline_length(self.measurement_w.mem_video)
            try:
                length = length / float(self.measurement_w.lePx.text()) * float(self.measurement_w.leRef.text())
            except Exception:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    "Check reference and pixel values! Values must be numeric.",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )

            append_results(
                self,
                (
                    n_player + 1,
                    media_file_name,
                    f"{self.getLaps():.03f}",
                    current_frame,
                    cfg.POLYGON_OBJECT,
                    cfg.NA,
                    cfg.NA,
                    round(length, 1),
                    round(area, 1),
                    cfg.NA,
                    str(self.measurement_w.mem_video),
                ),
            )

            self.measurement_w.flag_saved = False
            self.measurement_w.mem_points, self.measurement_w.mem_video = [], []

    # polyline
    elif self.measurement_w.rb_polyline.isChecked():
        if event.button() == Qt.LeftButton:
            draw_point(self, pixmap_x, pixmap_y, self.measurement_w.mark_color, n_player)
            if len(self.measurement_w.mem_points):
                draw_line(
                    self,
                    self.measurement_w.mem_points[-1][0],
                    self.measurement_w.mem_points[-1][1],
                    pixmap_x,
                    pixmap_y,
                    self.measurement_w.mark_color,
                    n_player,
                )
            self.measurement_w.mem_points.append((pixmap_x, pixmap_y))
            self.measurement_w.mem_video.append((x_video, y_video))

        if event.button() == Qt.RightButton:
            if current_frame not in self.measurement_w.draw_mem:
                self.measurement_w.draw_mem[current_frame] = []

            self.measurement_w.draw_mem[current_frame].append(
                {
                    "player": n_player,
                    "object_type": cfg.POLYLINE_OBJECT,
                    "color": self.measurement_w.mark_color,
                    "coordinates": self.measurement_w.mem_video,
                }
            )

            length = util.polyline_length(self.measurement_w.mem_video)
            try:
                length = length / float(self.measurement_w.lePx.text()) * float(self.measurement_w.leRef.text())
            except Exception:
                QMessageBox.critical(
                    None,
                    cfg.programName,
                    "Check reference and pixel values! Values must be numeric.",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )

            append_results(
                self,
                (
                    n_player + 1,
                    media_file_name,
                    f"{self.getLaps():.03f}",
                    current_frame,
                    cfg.POLYLINE_OBJECT,
                    cfg.NA,
                    cfg.NA,
                    round(length, 1),
                    cfg.NA,
                    cfg.NA,
                    str(self.measurement_w.mem_video),
                ),
            )
            self.measurement_w.mem_points, self.measurement_w.mem_video = [], []
            self.measurement_w.flag_saved = False

    else:
        self.measurement_w.status_lb.setText("<b>Choose a measurement type!</b>")


def redraw_measurements(self):
    """
    redraw measurements from previous frames
    """

    def scale_coord(coord_list: list) -> List[int]:
        """
        scale coordinates from original media resolution to pixmap
        """

        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
            # pixmap_size = QPixmap(self.images_list[self.image_idx]).size()
            original_width, original_height = self.current_image_size
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
            original_width = dw.player.width
            original_height = dw.player.height

        pixmap_coord_list: List[int] = []
        for idx, coord in enumerate(coord_list):
            if idx % 2 == 0:
                coord_pixmap = round(coord / original_width * dw.frame_viewer.pixmap().width())
            else:
                coord_pixmap = round(coord / original_height * dw.frame_viewer.pixmap().height())
            pixmap_coord_list.append(coord_pixmap)

        return pixmap_coord_list

    logging.debug("Redraw measurement marks")

    if not (hasattr(self, "measurement_w") and self.measurement_w is not None and self.measurement_w.isVisible()):
        return

    if not self.measurement_w.cbPersistentMeasurements.isChecked():
        self.measurement_w.draw_mem = {}
        return

    for idx, dw in enumerate(self.dw_player):
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.MEDIA:
            if dw.player.estimated_frame_number is not None:
                current_frame = dw.player.estimated_frame_number
            else:
                current_frame = cfg.NA
        if self.pj[cfg.OBSERVATIONS][self.observationId][cfg.TYPE] == cfg.IMAGES:
            current_frame = self.image_idx

        for frame in self.measurement_w.draw_mem:
            for element in self.measurement_w.draw_mem[frame]:
                if frame == current_frame:
                    elements_color = element["color"]
                else:
                    elements_color = cfg.PASSIVE_MEASUREMENTS_COLOR

                if element["player"] == idx:
                    if element["object_type"] == cfg.POINT_OBJECT:
                        x, y = scale_coord(element["coordinates"][0])
                        draw_point(self, int(x), int(y), elements_color, n_player=idx)

                    if element["object_type"] == cfg.SEGMENT_OBJECT:
                        x1, y1 = scale_coord(element["coordinates"][0])
                        x2, y2 = scale_coord(element["coordinates"][1])
                        draw_line(self, x1, y1, x2, y2, elements_color, n_player=idx)
                        draw_point(self, x1, y1, elements_color, n_player=idx)
                        draw_point(self, x2, y2, elements_color, n_player=idx)

                    if element["object_type"] in (cfg.ANGLE_OBJECT, cfg.ORIENTED_ANGLE_OBJECT):
                        x1, y1 = scale_coord(element["coordinates"][0])
                        x2, y2 = scale_coord(element["coordinates"][1])
                        x3, y3 = scale_coord(element["coordinates"][2])
                        draw_line(self, x1, y1, x2, y2, elements_color, n_player=idx)
                        draw_line(self, x1, y1, x3, y3, elements_color, n_player=idx)
                        draw_point(self, x1, y1, elements_color, n_player=idx)
                        draw_point(self, x2, y2, elements_color, n_player=idx)
                        draw_point(self, x3, y3, elements_color, n_player=idx)

                    if element["object_type"] == cfg.POLYGON_OBJECT:
                        polygon = QPolygon()
                        for x, y in element["coordinates"]:
                            x, y = scale_coord([x, y])
                            polygon.append(QPoint(x, y))
                        painter = QPainter()
                        painter.begin(self.dw_player[idx].frame_viewer.pixmap())
                        painter.setPen(QColor(elements_color))
                        painter.drawPolygon(polygon)
                        painter.end()
                        dw.frame_viewer.update()

                    if element["object_type"] == cfg.POLYLINE_OBJECT:
                        for idx1, p1 in enumerate(element["coordinates"][:-1]):
                            x1, y1 = scale_coord(p1)
                            p2 = element["coordinates"][idx1 + 1]
                            x2, y2 = scale_coord(p2)

                            draw_line(self, x1, y1, x2, y2, elements_color, n_player=idx)


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    w = wgMeasurement(logging.getLogger().getEffectiveLevel())
    w.show()

    sys.exit(app.exec_())



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.



Plot events in real time
"""

import matplotlib

matplotlib.use("Qt5Agg")

import numpy as np
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QLabel
from PyQt5.QtCore import pyqtSignal, QEvent
from PyQt5 import Qt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from matplotlib.figure import Figure

from . import config as cfg

# matplotlib.pyplot.switch_backend("Qt5Agg")


class Plot_events_RT(QWidget):
    # send keypress event to mainwindow
    sendEvent = pyqtSignal(QEvent)

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Events plot")

        self.interval = 60  # default interval of visualization (in seconds)
        self.time_mem = -1

        self.events_mem = {"init": 0}

        self.cursor_color = cfg.REALTIME_PLOT_CURSOR_COLOR  # default cursor color
        self.observation_type = cfg.MEDIA
        self.groupby = "behaviors"  # group results by "behaviors" or "modifiers"

        self.figure = Figure()
        self.ax = self.figure.add_subplot(1, 1, 1)

        self.canvas = FigureCanvas(self.figure)

        layout = QVBoxLayout()
        layout.addWidget(self.canvas)

        hlayout1 = QHBoxLayout()
        hlayout1.addWidget(QLabel("Time interval"))
        hlayout1.addWidget(QPushButton("+", self, clicked=lambda: self.time_interval_changed(1), focusPolicy=Qt.Qt.NoFocus))
        hlayout1.addWidget(QPushButton("-", self, clicked=lambda: self.time_interval_changed(-1), focusPolicy=Qt.Qt.NoFocus))
        self.pb_mode = QPushButton("Include modifiers", self, clicked=self.change_mode, focusPolicy=Qt.Qt.NoFocus)
        hlayout1.addWidget(self.pb_mode)
        layout.addLayout(hlayout1)

        self.setLayout(layout)

        self.installEventFilter(self)

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.sendEvent.emit(event)
            return True
        else:
            return False

    def change_mode(self) -> None:
        """
        Change plot mode
        "behaviors" -> plot behaviors without modifiers
        "modifiers" -> plot behaviors and modifiers
        """

        if self.groupby == "behaviors":
            self.groupby = "modifiers"
            self.pb_mode.setText("Show behaviors w/o modifiers")
        else:
            self.groupby = "behaviors"
            self.pb_mode.setText("Include modifiers")

    def time_interval_changed(self, action: int) -> None:
        """
        change the time interval for events plot

        Args:
            action (int): -1 decrease time interval, +1 increase time interval

        Returns:
            None
        """

        if action == -1 and self.interval <= 5:
            return
        self.interval += 5 * action
        self.plot_events(current_time=self.time_mem, force_plot=True)

    def aggregate_events(self, events: list, start: float, end: float) -> dict:
        """
        aggregate state events
        take consideration of subject and modifiers

        Args:
            events (list): list of events
            start (float): initial value
            end (float): final value

        Returns:
            dict

        """

        def group(subject: str, code: str, modifier: str) -> tuple:
            if self.groupby == "behaviors":
                return (subject, code)
            else:  # with modifiers
                return (subject, code, modifier)

        mem_behav = {}
        intervals_behav = {}

        for event in events:
            intervals_behav[group(event[1], event[2], event[3])] = [(0, 0)]

        for event in events:
            time_, subject, code, modifier = event[:4]
            key = group(subject, code, modifier)

            # check if code is state
            if code in self.state_events_list:
                if key in mem_behav and mem_behav[key] is not None:
                    # stop interval

                    # check if event is in interval start-end
                    if any(
                        (
                            start <= mem_behav[key] <= end,
                            start <= time_ <= end,
                            mem_behav[key] <= start and time_ > end,
                        )
                    ):
                        intervals_behav[key].append((float(mem_behav[key]), float(time_)))
                    mem_behav[key] = None
                else:
                    # start interval
                    mem_behav[key] = time_

            else:  # point event
                if start <= time_ <= end:
                    intervals_behav[key].append((float(time_), float(time_) + self.point_event_plot_duration * 50))  # point event -> 1 s

        # check if intervals are closed
        for k in mem_behav:
            if mem_behav[k] is not None:  # interval open
                if self.observation_type == cfg.LIVE:
                    intervals_behav[k].append((float(mem_behav[k]), float((end + start) / 2)))  # close interval with current time

                elif self.observation_type == cfg.MEDIA:
                    intervals_behav[k].append((float(mem_behav[k]), float(end)))  # close interval with end value

        return intervals_behav

    def plot_events(self, current_time: float, force_plot: bool = False):
        """
        plot events centered on the current time

        Args:
            current_time (float): time for displaying events
            force_plot (bool): force plot even if media paused
        """

        self.events = self.aggregate_events(self.events_list, current_time - self.interval / 2, current_time + self.interval / 2)

        if not force_plot and current_time == self.time_mem:
            return

        self.time_mem = current_time

        if self.events != self.events_mem:
            left, duration = {}, {}
            for k in self.events:
                left[k] = []
                duration[k] = []
                for interv in self.events[k]:
                    left[k].append(interv[0])
                    duration[k].append(interv[1] - interv[0])

            self.behaviors, self.durations, self.lefts, self.colors = [], [], [], []
            for k in self.events:
                if self.groupby == "behaviors":
                    subject_name, bevavior_code = k
                    if subject_name == "":
                        subject_name = "No focal"
                    behav_col = self.behav_color[bevavior_code]
                    self.behaviors.extend([f"{subject_name} - {bevavior_code}"] * len(self.events[k]))
                    self.colors.extend([behav_col] * len(self.events[k]))
                else:  # with modifiers
                    subject_name, bevavior_code, modifier = k
                    behav_col = self.behav_color[bevavior_code]
                    self.behaviors.extend([f"{subject_name} - {bevavior_code} ({modifier})"] * len(self.events[k]))
                    self.colors.extend([behav_col] * len(self.events[k]))

                self.lefts.extend(left[k])
                self.durations.extend(duration[k])

            self.events_mem = self.events

        self.ax.clear()
        self.ax.set_xlim(current_time - self.interval / 2, current_time + self.interval / 2)

        self.ax.axvline(x=current_time, color=self.cursor_color, linestyle="-")

        self.ax.barh(y=np.array(self.behaviors), width=self.durations, left=self.lefts, color=self.colors, height=0.5)

        self.canvas.draw()
        self.figure.canvas.flush_events()



"""
module for testing otx_parser.py

https://realpython.com/python-continuous-integration/

pytest -s -vv test_otx_parser.py
"""

import os
import pytest
import sys
import json

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from boris import otx_parser

@pytest.fixture()
def before():
    os.system("rm -rf output")
    os.system("mkdir output")

class Test_otx_to_boris(object):

    def test_otx(self):
        boris_project = otx_parser.otx_to_boris("files/otx_parser_test.otx")
        pj = json.loads(open("files/otx_import_test.boris").read())
        assert boris_project == pj

        #with open("1", "w") as f:
        #    f.write(json.dumps(boris_project))


    def test_otb(self):
        boris_project = otx_parser.otx_to_boris("files/otx_parser_test.otb")
        pj = json.loads(open("files/otx_import_test.boris").read())
        assert boris_project == pj
        '''
        with open("1", "w") as f:
            f.write(json.dumps(boris_project))

        '''




"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.
"""

import pathlib
import re
import statistics
import sys

import tablib
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QMessageBox,
    QPushButton,
    QRadioButton,
    QSizePolicy,
    QSpacerItem,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
)

from . import config as cfg
from . import db_functions, dialog, observation_operations
from . import portion as Interval
from . import project_functions, select_observations, select_subj_behav
from . import utilities as util


def icc(i: list):
    """
    create a closed-closed interval
    """
    return Interval.closed(i[0], i[1])


def ico(i: list):
    """
    create a closed-open interval
    """
    return Interval.closedopen(i[0], i[1])


def io(i: list):
    """
    create a open interval
    """
    return Interval.open(i[0], i[1])


class Advanced_event_filtering_dialog(QDialog):
    """
    Dialog for visualizing advanced event filtering results
    """

    summary_header: tuple = (
        "Observation id",
        "Number of occurences",
        "Total duration (s)",
        "Duration mean (s)",
        "Std Dev",
    )
    details_header: tuple = ("Observation id", "Comment", "Start time", "Stop time", "Duration (s)")

    def __init__(self, events):
        super().__init__()

        self.events = events
        self.out = []
        self.setWindowTitle("Advanced event filtering")

        vbox = QVBoxLayout()

        self.lb_time_interval = QLabel()
        vbox.addWidget(self.lb_time_interval)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Filter"))
        self.logic = QLineEdit("")
        hbox.addWidget(self.logic)
        self.pb_filter = QPushButton("Filter events", clicked=self.filter)
        hbox.addWidget(self.pb_filter)
        self.pb_clear = QPushButton("Clear", clicked=self.logic.clear)
        self.pb_clear.setIcon(QIcon.fromTheme("edit-clear"))
        hbox.addWidget(self.pb_clear)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        self.rb_summary = QRadioButton("Summary", toggled=self.filter)
        self.rb_summary.setChecked(True)
        hbox.addWidget(self.rb_summary)
        self.rb_details = QRadioButton("Details", toggled=self.filter)
        hbox.addWidget(self.rb_details)
        vbox.addLayout(hbox)

        hbox = QHBoxLayout()
        vbox2 = QVBoxLayout()
        vbox2.addWidget(QLabel("Subjects"))
        self.lw1 = QListWidget()
        vbox2.addWidget(self.lw1)
        hbox.addLayout(vbox2)

        vbox2 = QVBoxLayout()
        vbox2.addWidget(QLabel("Behaviors"))
        self.lw2 = QListWidget()
        vbox2.addWidget(self.lw2)
        hbox.addLayout(vbox2)
        self.add_subj_behav_button = QPushButton("", clicked=self.add_subj_behav)
        self.add_subj_behav_button.setIcon(QIcon.fromTheme("go-up"))
        hbox.addWidget(self.add_subj_behav_button)

        vbox2 = QVBoxLayout()
        vbox2.addWidget(QLabel("Logical operators"))
        self.lw3 = QListWidget()
        self.lw3.addItems(["AND", "OR"])
        vbox2.addWidget(self.lw3)
        hbox.addLayout(vbox2)
        self.add_logic_button = QPushButton("", clicked=self.add_logic)
        self.add_logic_button.setIcon(QIcon.fromTheme("go-up"))
        hbox.addWidget(self.add_logic_button)

        vbox.addLayout(hbox)

        self.lb_results = QLabel("Results")
        vbox.addWidget(self.lb_results)

        self.tw = QTableWidget(self)
        vbox.addWidget(self.tw)

        hbox = QHBoxLayout()
        hbox.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        self.pb_save = QPushButton("Save results", clicked=self.save_results)
        hbox.addWidget(self.pb_save)
        self.pb_close = QPushButton("Close", clicked=self.close)
        hbox.addWidget(self.pb_close)
        vbox.addLayout(hbox)

        self.setLayout(vbox)

        subjects_list, behaviors_list = [], []
        for obs_id in events:
            for subj_behav in events[obs_id]:
                subj, behav = subj_behav.split("|")
                subjects_list.append(subj)
                behaviors_list.append(behav)
        subjects_set = sorted(set(subjects_list))
        behaviors_set = sorted(set(behaviors_list))

        self.lw1.addItems(subjects_set)
        self.lw2.addItems(behaviors_set)

        self.resize(640, 640)

    def add_subj_behav(self):
        """
        add subject|behavior of selected listwidgets items in lineEdit
        """
        if self.lw1.currentItem() and self.lw2.currentItem():
            self.logic.insert(f'"{self.lw1.currentItem().text()}|{self.lw2.currentItem().text()}" ')
        else:
            QMessageBox.warning(self, cfg.programName, "Select a subject and a behavior")

    def add_logic(self):
        """
        add selected logic operaton to lineedit
        """
        if self.lw3.currentItem():
            text = ""
            if self.lw3.currentItem().text() == "AND":
                text = " & "
            if self.lw3.currentItem().text() == "OR":
                text = " | "
            if text:
                self.logic.insert(text)
        else:
            QMessageBox.warning(self, cfg.programName, "Select a logical operator")

    def filter(self):
        """
        filter events
        """
        if not self.logic.text():
            return
        if self.logic.text().count('"') % 2:
            QMessageBox.warning(self, cfg.programName, 'Wrong number of double quotes (")')
            return

        sb_list = re.findall('"([^"]*)"', self.logic.text())

        self.out = []
        flag_error = False
        for obs_id in self.events:
            logic = self.logic.text()
            for sb in set(sb_list):
                logic = logic.replace(f'"{sb}"', f'self.events[obs_id]["{sb}"]')
                if sb not in self.events[obs_id]:
                    self.events[obs_id][sb] = io([0, 0])

            try:
                eval_result = eval(logic)
                for i in eval_result:
                    if not i.empty:
                        self.out.append([obs_id, "", f"{i.lower}", f"{i.upper}", f"{i.upper - i.lower:.3f}"])
            except KeyError:
                self.out.append([obs_id, "subject / behavior not found", cfg.NA, cfg.NA, cfg.NA])
            except Exception:
                error_type, _, _ = util.error_info(sys.exc_info())
                self.out.append([obs_id, f"Error in {self.logic.text()}: {error_type} ", cfg.NA, cfg.NA, cfg.NA])
                flag_error = True

        self.tw.clear()

        if flag_error or self.rb_details.isChecked():
            self.lb_results.setText(f"Results ({len(self.out)} event{'s'*(len(self.out) > 1)})")

            self.tw.setRowCount(len(self.out))
            self.tw.setColumnCount(len(self.details_header))  # obs_id, comment, start, stop, duration
            self.tw.setHorizontalHeaderLabels(self.details_header)

        if not flag_error and self.rb_summary.isChecked():
            summary = {}
            for row in self.out:
                obs_id, _, start, stop, duration = row
                if obs_id not in summary:
                    summary[obs_id] = []
                summary[obs_id].append(float(duration))

            self.out = []
            for obs_id in summary:
                self.out.append(
                    [
                        obs_id,
                        str(len(summary[obs_id])),
                        str(round(sum(summary[obs_id]), 3)),
                        str(round(statistics.mean(summary[obs_id]), 3)),
                        str(round(statistics.stdev(summary[obs_id]), 3)) if len(summary[obs_id]) > 1 else "NA",
                    ]
                )

            self.lb_results.setText(f"Results ({len(summary)} observation{'s'*(len(summary) > 1)})")
            self.tw.setRowCount(len(summary))
            self.tw.setColumnCount(len(self.summary_header))  # obs_id, mean, stdev
            self.tw.setHorizontalHeaderLabels(self.summary_header)

        for r in range(len(self.out)):
            for c in range(self.tw.columnCount()):
                item = QTableWidgetItem()
                item.setText(self.out[r][c])
                item.setFlags(Qt.ItemIsEnabled)
                self.tw.setItem(r, c, item)

    def save_results(self):
        """
        save results
        """

        file_formats = [
            cfg.TSV,
            cfg.CSV,
            cfg.ODS,
            cfg.XLSX,
            cfg.XLS,
            cfg.HTML,
        ]

        file_name, filter_ = QFileDialog().getSaveFileName(None, "Save results", "", ";;".join(file_formats))
        if not file_name:
            return

        output_format = cfg.FILE_NAME_SUFFIX[filter_]

        if pathlib.Path(file_name).suffix != "." + output_format:
            file_name = str(pathlib.Path(file_name)) + "." + output_format
            # check if file with new extension already exists
            if pathlib.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                    == cfg.CANCEL
                ):
                    return

        if self.rb_details.isChecked():
            tablib_dataset = tablib.Dataset(headers=self.details_header)
        if self.rb_summary.isChecked():
            tablib_dataset = tablib.Dataset(headers=self.summary_header)
        tablib_dataset.title = util.safe_xl_worksheet_title(self.logic.text(), output_format)

        [tablib_dataset.append(x) for x in self.out]

        try:
            with open(file_name, "wb") as f:
                if filter_ in (cfg.TSV, cfg.CSV, cfg.HTML):
                    f.write(str.encode(tablib_dataset.export(output_format)))
                if filter_ in (cfg.ODS, cfg.XLSX, cfg.XLS):
                    f.write(tablib_dataset.export(output_format))

        except Exception:
            QMessageBox.critical(self, cfg.programName, f"The file {file_name} can not be saved")


def event_filtering(self):
    """
    advanced event filtering
    the portion module is used to do operations on intervals (intersection, union)
    """

    _, selected_observations = select_observations.select_observations2(
        self, cfg.MULTIPLE, "Select observations for advanced event filtering"
    )
    if not selected_observations:
        return

    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)
    # exit with message if events do not have timestamp
    if start_coding.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have a timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    max_media_duration_all_obs, _ = observation_operations.media_duration(self.pj[cfg.OBSERVATIONS], selected_observations)

    parameters = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=start_coding,
        end_coding=end_coding,
        maxTime=max_media_duration_all_obs,
        flagShowIncludeModifiers=False,
        flagShowExcludeBehaviorsWoEvents=False,
        by_category=False,
        n_observations=len(selected_observations),
    )
    if parameters == {}:
        return

    if not parameters[cfg.SELECTED_SUBJECTS] or not parameters[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    _, _, db_connector = db_functions.load_aggregated_events_in_db(
        self.pj, parameters[cfg.SELECTED_SUBJECTS], selected_observations, parameters[cfg.SELECTED_BEHAVIORS]
    )

    cursor = db_connector.cursor()

    if parameters[cfg.TIME_INTERVAL] in (cfg.TIME_EVENTS, cfg.TIME_FULL_OBS):
        cursor.execute("SELECT MIN(start), MAX(stop) FROM aggregated_events")
        min_time, max_time = cursor.fetchone()

    if parameters[cfg.TIME_INTERVAL] == cfg.TIME_ARBITRARY_INTERVAL:
        min_time = float(parameters[cfg.START_TIME])
        max_time = float(parameters[cfg.END_TIME])

    cursor.execute(
        "UPDATE aggregated_events SET start = ? WHERE start < ? AND stop BETWEEN ? AND ?",
        (
            min_time,
            min_time,
            min_time,
            max_time,
        ),
    )
    cursor.execute(
        "UPDATE aggregated_events SET stop = ? WHERE stop > ? AND start BETWEEN ? AND ?",
        (
            max_time,
            max_time,
            min_time,
            max_time,
        ),
    )
    cursor.execute(
        "UPDATE aggregated_events SET start = ?, stop = ? WHERE start < ? AND stop > ?",
        (
            min_time,
            max_time,
            min_time,
            max_time,
        ),
    )

    cursor.execute(
        "DELETE FROM aggregated_events WHERE (start < ? AND stop < ?) OR (start > ? AND stop > ?)",
        (
            min_time,
            min_time,
            max_time,
            max_time,
        ),
    )

    # create intervals from DB
    cursor.execute("SELECT observation, subject, behavior, start, stop FROM aggregated_events")

    events: dict = {}
    for row in cursor.fetchall():
        obs, subj, behav, start, stop = row
        if obs not in events:
            events[obs]: dict = {}

        # use function in base at event (state or point)
        interval_func = icc if start == stop else ico

        if f"{subj}|{behav}" not in events[obs]:
            # create new interval
            events[obs][f"{subj}|{behav}"] = interval_func([start, stop])
        else:
            # append to existing interval
            events[obs][f"{subj}|{behav}"] |= interval_func([start, stop])

    w = Advanced_event_filtering_dialog(events)
    w.lb_time_interval.setText(
        ("Time interval: " f"{util.smart_time_format(min_time, self.timeFormat)} - " f"{util.smart_time_format(max_time, self.timeFormat)}")
    )
    w.exec_()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2024 Olivier Friard


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
  MA 02110-1301, USA.

"""

import wave
import matplotlib

matplotlib.use("Qt5Agg")

import numpy as np

from . import config as cfg

from PyQt5.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QLabel,
    QSpinBox,
)
from PyQt5.QtCore import pyqtSignal, QEvent
from PyQt5 import Qt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.ticker as mticker

# matplotlib.pyplot.switch_backend("Qt5Agg")


class Plot_spectrogram_RT(QWidget):
    # send keypress event to mainwindow
    sendEvent = pyqtSignal(QEvent)

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Spectrogram")

        self.interval = 10  # interval of visualization (in seconds)
        self.time_mem = -1

        self.cursor_color = cfg.REALTIME_PLOT_CURSOR_COLOR

        self.spectro_color_map = matplotlib.pyplot.get_cmap("viridis")

        self.figure = Figure()
        self.ax = self.figure.add_subplot(1, 1, 1)

        self.canvas = FigureCanvas(self.figure)

        layout = QVBoxLayout()
        layout.addWidget(self.canvas)

        hlayout1 = QHBoxLayout()
        hlayout1.addWidget(QLabel("Time interval"))
        hlayout1.addWidget(
            QPushButton(
                "+",
                self,
                clicked=lambda: self.time_interval_changed(1),
                focusPolicy=Qt.Qt.NoFocus,
            )
        )
        hlayout1.addWidget(
            QPushButton(
                "-",
                self,
                clicked=lambda: self.time_interval_changed(-1),
                focusPolicy=Qt.Qt.NoFocus,
            )
        )
        layout.addLayout(hlayout1)

        hlayout2 = QHBoxLayout()
        hlayout2.addWidget(QLabel("Frequency interval"))
        self.sb_freq_min = QSpinBox(valueChanged=self.frequency_interval_changed, focusPolicy=Qt.Qt.NoFocus)
        self.sb_freq_min.setRange(0, 200000)
        self.sb_freq_min.setSingleStep(100)
        self.sb_freq_max = QSpinBox(valueChanged=self.frequency_interval_changed, focusPolicy=Qt.Qt.NoFocus)
        self.sb_freq_max.setRange(0, 200000)
        self.sb_freq_max.setSingleStep(100)
        hlayout2.addWidget(self.sb_freq_min)
        hlayout2.addWidget(self.sb_freq_max)
        layout.addLayout(hlayout2)

        self.setLayout(layout)

        self.installEventFilter(self)

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            self.sendEvent.emit(event)
            return True
        else:
            return False

    def get_wav_info(self, wav_file: str):
        """
        read wav file and extract information

        Args:
            wav_file (str): path of wav file

        Returns:
            np.array: signal contained in wav file
            int: frame rate of wav file

        """

        try:
            wav = wave.open(wav_file, "r")
            frames = wav.readframes(-1)
            sound_info = np.fromstring(frames, dtype=np.int16)
            frame_rate = wav.getframerate()
            wav.close()
            return sound_info, frame_rate
        except Exception:
            return np.array([]), 0

    def time_interval_changed(self, action: int):
        """
        change the time interval for plotting spectrogram

        Args:
            action (int): -1 decrease time interval, +1 increase time interval

        Returns:
            None
        """

        if action == -1 and self.interval <= 5:
            return
        self.interval += 5 * action
        self.plot_spectro(current_time=self.time_mem, force_plot=True)

    def frequency_interval_changed(self):
        """
        change the frequency interval for plotting spectrogram
        """
        self.plot_spectro(current_time=self.time_mem, force_plot=True)

    def load_wav(self, wav_file_path: str) -> dict:
        """
        load wav file in numpy array

        Args:
            wav_file_path (str): path of wav file

        Returns:
            dict: "error" key if error, "media_length" and "frame_rate"
        """

        try:
            self.sound_info, self.frame_rate = self.get_wav_info(wav_file_path)
            if not self.frame_rate:
                return {"error": f"unknown format for file {wav_file_path}"}
        except FileNotFoundError:
            return {"error": f"File not found: {wav_file_path}"}

        self.media_length = len(self.sound_info) / self.frame_rate

        self.wav_file_path = wav_file_path

        return {"media_length": self.media_length, "frame_rate": self.frame_rate}

    def plot_spectro(self, current_time: float, force_plot: bool = False):
        """
        plot sound spectrogram centered on the current time

        Args:
            current_time (float): time for displaying spectrogram
            force_plot (bool): force plot even if media paused
        """

        if not force_plot and current_time == self.time_mem:
            return

        self.time_mem = current_time

        self.ax.clear()

        # start
        if current_time <= self.interval / 2:
            self.ax.specgram(
                self.sound_info[: int((self.interval) * self.frame_rate)],
                mode="psd",
                # NFFT=1024,
                Fs=self.frame_rate,
                # noverlap=900,
                cmap=self.spectro_color_map,
            )

            self.ax.set_xlim(current_time - self.interval / 2, current_time + self.interval / 2)

            # cursor
            self.ax.axvline(x=current_time, color=self.cursor_color, linestyle="-")

        elif current_time >= self.media_length - self.interval / 2:
            i = int(round(len(self.sound_info) - (self.interval * self.frame_rate), 0))

            self.ax.specgram(
                self.sound_info[i:],
                mode="psd",
                # NFFT=1024,
                Fs=self.frame_rate,
                # noverlap=900,
                cmap=self.spectro_color_map,
            )

            lim1 = current_time - (self.media_length - self.interval / 2)
            lim2 = lim1 + self.interval

            self.ax.set_xlim(lim1, lim2)

            self.ax.xaxis.set_major_locator(mticker.FixedLocator(self.ax.get_xticks().tolist()))
            self.ax.set_xticklabels([str(round(w + self.media_length - self.interval, 1)) for w in self.ax.get_xticks()])

            # cursor
            self.ax.axvline(x=lim1 + self.interval / 2, color=self.cursor_color, linestyle="-")

        # middle
        else:
            self.ax.specgram(
                self.sound_info[
                    int(round((current_time - self.interval / 2) * self.frame_rate, 0)) : int(
                        round((current_time + self.interval / 2) * self.frame_rate, 0)
                    )
                ],
                mode="psd",
                # NFFT=1024,
                Fs=self.frame_rate,
                # noverlap=900,
                cmap=self.spectro_color_map,
            )

            self.ax.xaxis.set_major_locator(mticker.FixedLocator(self.ax.get_xticks().tolist()))
            self.ax.set_xticklabels([str(round(current_time + w - self.interval / 2, 1)) for w in self.ax.get_xticks()])

            # cursor
            self.ax.axvline(x=self.interval / 2, color=self.cursor_color, linestyle="-")

        self.ax.set_ylim(self.sb_freq_min.value(), self.sb_freq_max.value())
        """self.figure.subplots_adjust(wspace=0, hspace=0)"""

        self.canvas.draw()
